/**********************/
/* read flash from j5 */
/**********************/


.text
.code 16
.globl _start

_start: b .main


/* send byte from r0 */
.sendbyte:
           ldr r1, s_stat
           ldrb r1, [r1]
           lsr r1, r1, #7
           bcc .sendbyte
           ldr r1, s_data
           strb r0, [r1]
           mov r1, #0

.wait_send:
            nop
            add r1, #1
            lsl r1, r1, #0x18
            lsr r1, r1, #0x18
            cmp r1, #0
            bne .wait_send
            bx LR

/* end of sendbyte */

/* receive byte and place it to r0 */
.receivebyte:
              ldr r1, s_stat
              ldrb r1, [r1]
              lsr r1, r1, #1
              bcc .receivebyte
              ldr r1, s_data
              ldrb r0, [r1]
              bx LR

/* end of receivebyte */


/******** main entry ********/

.main:
       push {LR}

        ldr r1, timer0excntrl /* disable watchdog */
        mov r3, #0xF5
        mov r4, #0xA0
        strb r3, [r1]
        strb r4, [r1]

.send_alive:
             mov r0, #0x3E
             bl .sendbyte
.wait_loop:
            bl .receivebyte
            cmp r0, #0x78 /* 'quit' command received ? */
            beq .end
            cmp r0, #0x3E /* 'are_you_alive' command received ? */
            beq .send_alive
            cmp r0, #0x64 /* 'send datablock' command received ? */
            beq .send_data
            cmp r0, #0x77 /* 'write byte' command received ? */
            beq .write_by
            b   .wait_loop /* if no known command received then jump to wait_loop */

.write_by:  bl .receivebyte /* receive the highest byte of address */
            lsl r2, r0, #0x18
            bl .receivebyte
            lsl r0, r0, #0x10
            add r2, r2, r0
            bl .receivebyte
            lsl r0, r0, #8
            add r2, r2, r0
            bl .receivebyte /* receive the lowest byte of address */
            add r2, r2, r0 /* r2 = begin address */
            mov r0, #0x57 /* send start_sentinel */
            bl .sendbyte
            bl .receivebyte
            strb r0, [r2]
            mov r0, #0x77 /* send end_write */
            bl .sendbyte
            b .wait_loop /* back to main loop */
            
.send_data: bl .receivebyte /* receive the highest byte of address */
            lsl r2, r0, #0x18
            bl .receivebyte
            lsl r0, r0, #0x10
            add r2, r2, r0
            bl .receivebyte
            lsl r0, r0, #8
            add r2, r2, r0
            bl .receivebyte /* receive the lowest byte of address */
            add r2, r2, r0 /* r2 = begin address */
            mov r0, #0x44 /* send start_sentinel */
            bl .sendbyte
            mov r3, #1 /* r3 = bytes to send */
            lsl r3, r3, #0x08 /* 256 bytes */
            mov r7, #0 /* clear 'checksum_register' */

.send_datablock:
            ldrb r0, [r2]
            add r7, r0
            bl .sendbyte
            add r2, #1 /* increment address */
            sub r3, #1 /* decrement bytes2send */
            bne .send_datablock /* data are sent, now send the end_sentinel and chksum */

            mov r0, #0x63 /* end sentinel */
            bl .sendbyte

            ldrh r6, mod
            and r7, r6 /* calc chksum */
            eor r7, r6

            lsr r0, r7, #0x08
            bl .sendbyte /* send hibyte of chksum */
            mov r0, r7
            bl .sendbyte /* send lobyte of chksum */
            b .wait_loop /* back to main loop */

.end:
            ldr r1, timer0excntrl /* enable watchdog */
            mov r3, #0x80
            eor r4, r4
            strb r3, [r1]
            strb r4, [r1]
            pop {PC}

.align 0
s_stat: .word 0xFFFF5805
s_data: .word 0xFFFF5800
timer0cntrl: .word 0xFFFFF800
timer0count: .word 0xFFFFF802
timer0excntrl: .word 0xFFFFF804
mod: .word 0x0000FFFF
